#!/usr/bin/env python

# change next line to debug=1 to debug script

debug=0

""" mdetails.cgi This file handles membership updates which a
    user can implement for themselves. 
    Richard Kay : 2010 - 2023
"""

if debug:
  import sys
  sys.stderr=sys.stdout
  print("Content-type: text/plain\n")

import cgiutils
import webinput
woi=webinput.Webinput()
import tablcgi
from table_details import mtab

v=cgiutils.get_vars('site.cfg')
dl=v['data_location']

def send_newpin(email,newpin):
  ''' send new pin to email address
  parameters:
  email - email address of applicant
  newpin - PIN user will need to login
  '''
  fromad=v['admin_e']
  regy=v['registry']
  url=v['homepage']
  message="From: "+regy+"  <"+v['admin_e']+">\n"
  to_line="To: %s\n" % email
  message+=to_line
  message+="Subject: "+regy+" your PIN has been reset \n\n"
  message+="Your new personal identification number (PIN) is %s .\n\n" % newpin
  message+="The accounts, membership and trading website is accessible\n"
  message+="at the following address:\n\n"+url
  cgiutils.send_mail(fromad,email,message)

def reset_pin(session,ac_id,members,midx):
  ''' resets the user PIN and emails new PIN to address
  for the user '''
  if ac_id not in members.keys(): # this really shouldnt happen
    error="Account id/no. doesn't exist. Can't reset PIN."
    cgiutils.html_end(session,error)
    return
  email=members.data[midx]['email']
  if cgiutils.is_email(email):
    newpin=cgiutils.make_pin(mini=10000000,maxi=99999999)
    members.data[midx]['pin']=newpin
    members.save()
    send_newpin(email,newpin)
    print("<p> PIN emailed for account: %s to address: %s </p>" % (ac_id,
         email))
  else: # this one is weird - how did they get here ?
    print('<p> As you do not have an email address registered there')
    print('is no way to send you a new PIN.</p>') 
  cgiutils.html_end(session,received=1)
  return

def get_address(session):
  address=woi.firstval("address")
  if not cgiutils.is_valid(address,max_leng=90,
      allowed='^[A-Za-z0-9][\w\'\-\,\.\r\n ]+$'):
    cgiutils.html_end(session,error="Invalid address.")
    return False
  return address

def get_postcode(session):
  postcode=woi.firstval("postcode")
  if postcode == '':
    return 'None'
  if not cgiutils.is_valid(postcode,max_leng=10, allowed='^[A-Za-z0-9][\w ]+$'):
    cgiutils.html_end(session,error="Invalid postcode.")
    return False
  return postcode

def get_district(session):
  district=woi.firstval("district")
  if district == '':
    return 'None'
  if not cgiutils.is_valid(district,max_leng=20, allowed='^[A-Za-z ]+$'):
    cgiutils.html_end(session,
       error="Invalid district use letters a-z and space only.")
    return False
  return district

def get_url(session):
  url=woi.firstval("url")
  if url=='' or url.lower() == 'none':
    return 'None' # null value is a valid replacement
  valid,error=cgiutils.is_url(url)
  if not valid:
    cgiutils.html_end(session,
       error="Invalid homepage/URL. The URL parser reported: "+error)
    return False
  # strip the url to remove any query string. We don't do XSS or CSRF
  safeurl=cgiutils.url_strip(url)
  return safeurl  

def modify(session,ac_id,members,midx):
  modifications=0
  auditstr='' # used to record changes
  if 'address' in woi.keys():
    oldaddress=members.data[midx]['address']
    address=get_address(session)
    if not address:
      return
    if address != 'None':
      modifications+=1
      members.data[midx]['address']=address
      auditstr+=' changed address from '+oldaddress+' to '+address
  if 'postcode' in woi.keys():
    oldpostcode=members.data[midx]['postcode']
    postcode=get_postcode(session)
    if not postcode:
      return
    if postcode != 'None':
      modifications+=1
      members.data[midx]['postcode']=postcode
      auditstr+=' changed postcode from '+oldpostcode+' to '+postcode
  if 'district' in woi.keys():
    olddistrict=members.data[midx]['district']
    district=get_district(session)
    if not district:
      return
    if district != 'None':
      modifications+=1
      members.data[midx]['district']=district
      auditstr+=' changed district from '+olddistrict+' to '+district
  if 'url' in woi.keys():
    oldurl=members.data[midx]['url']
    url=get_url(session)
    if not url:
      return
    # check if provisionals allowed to update URL
    status=members.data[midx]['active']
    if ( v['provisional_can_url'].upper() == 'N' and 
        status == 'P' ) : 
      errmes='Site policy prevents provisional members linking a homepage.'
      cgiutils.html_end(session,error=errmes)
      return
    if url != 'None':
      modifications+=1
      members.data[midx]['url']=url
      auditstr+=' changed url from '+oldurl+' to '+url
  if modifications == 0:
    errmes='You have not input any changed details.</p>'
    cgiutils.html_end(session,error=errmes)
    return 
  else:
    members.save()
    print('<p> %d field/s modified and changes logged.</p>' % modifications)
    # audit log the change - a member changing own street address details doesnt
    # need further authorisation, but we log it in case of misuse
    # so that a permanent record of the changes made is kept
    cgiutils.auth(session,
         ac_id,"MEMBMOD",auditstr,whom=ac_id,auditonly=True)
    cgiutils.html_end(session,received=1)
    return

def main():
  session=False
  cgiutils.html_header(title="LETS Member Contact Directory Access",
                       bgcolor='"#FFDDFF"')
  valid,ac_id,pincookie,ua=cgiutils.get_hidden(woi)
  if not valid:
    return # error message already sent
  valid,option=cgiutils.get_option(woi,buttons=
     ["rpin","view","modify"])
  if not valid:
    return # error message already sent
  # check pin_or_cookie and account
  session=cgiutils.login(ac_id,pincookie,ua)
  valid=session.login()
  cookie=session.cookie
  if not valid:
    cgiutils.html_header(title="LETS Accounting and Directory System")
    cgiutils.html_end(session,
       error="Invalid or stale or logged out session.")
    return
  # 
  members=tablcgi.table(mtab,dir=dl)
  midx=members.find(ac_id)
  if option == "rpin":
    reset_pin(session,ac_id,members,midx)
    return
  elif option == "view":
    mvd=tablcgi.table_def(mtab) # clone member view table details
    mvd.file=None
    mtv=tablcgi.table(mvd) # member table view
    row=members.data[midx] 
    mtv.addrow(row)
    mtv.tab2html(skip_cols=['pin'])
    cgiutils.html_end(session)
    return
  elif option == "modify":
    modify(session,ac_id,members,midx)
    return

if __name__ == "__main__":
  try: 
    main()
  except:
    import traceback
    print("error detected in mdetails.cgi main()")
    traceback.print_exc()

