#!/usr/bin/env python

''' monthly_charges.py
    PyLETS client program which computes and applies routine e.g. 
    monthly charges to all active accounts. 

    This application is configured by editing the charges.cfg
    file which is within the pylets configuration directory,
    typically /etc/pylets . The configuration file controls
    the account ids which don't pay charges, the amount, and
    description of the payment, and the system account into which
    payments are made.    
 
    Each payment is compiled based on the members table.
    The payment is made by this application acting as 
    an HTTP client, which ensures that the normal and 
    tested payment code is used in all payment events. 
 
    Richard Kay, 30 Oct 2009
'''

# interactive - change to True to make program stepped and noisy
interactive=False

import cgiutils
import tablcgi
from table_details import mtab
site_vars=cgiutils.get_vars('site.cfg')
charge_vars=cgiutils.get_vars('charges.cfg')
dl=site_vars['data_location']

def nochargeacs():
  ''' returns list of account names which are not charged, read
  from the no_charge text file, one account per line '''
  acs=charge_vars['nocharge_acs'] 
  ncas=acs.split(',') # split comma seperated string into list
  return ncas

def do_post(postdata):
  ''' perform a single web request, by doing a HTTP post
  operation on CGI server, and reporting pass/fail .
  This is based upon whether the result string
  contains expected string (for a pass) or not (for a fail)
  '''
  import urllib.request, urllib.error, urllib.parse
  req = urllib.request.Request(url=postdata['url'],data=postdata['data'])
  fr = urllib.request.urlopen(req)
  result=fr.read()
  if result.find(postdata['expected']) >= 0:
    # payment succeeded if expected is substring of result
    print(('processed record: '+str(postdata['recno'])+' '
          +postdata['description']))
  else:
    print(('failed to process record: '+str(postdata['recno'])+
           ' '+postdata['description']))
    print('response: ' + result)

def prepare_post(recno,admin_pin,domain,webapp_path,
                 payer,payer_name,payto,payto_name,amount,desc):
  # replacement tuple to interpolate into postdata
  rtuple=(admin_pin, 
          payer, 
          payer_name, 
          payto, 
          payto_name,
          amount, 
          desc        )
  datastr='ac_id=admin&pin=%s&ua=admin&from_id=%s&from_name=%s&to_id=%s&to_name=%s&amount=%s&details=%s' 
  postdata={'url':'http://'+domain+webapp_path+'ack.cgi',
            'data':datastr % rtuple, 
            'expected':"successfully input",
            'recno':recno,
            'description':'system charges payment by %s' % payer }
  return postdata

def read_active():
  ''' Gets and returns a list of active member account IDs and names.  
  '''
  actives=[]
  ncas=nochargeacs() # get list of non charge paying accounts
  members=tablcgi.table(mtab,dir=dl)
  for mem in members.data:
    if mem["active"] == 'Y' and not mem['ac_id'] in ncas:
      actives.append({'payer':mem['ac_id'],'payer_name':mem['name']})
  return actives

def get_admin_pin():
  ''' Gets and returns admin account PIN used to apply charges '''
  members=tablcgi.table(mtab,dir=dl)
  return str(members.data[members.find('admin')]['pin']) 

def main():
  import getpass
  domain=charge_vars['domain']
  webapp_path=charge_vars['dir']
  amount=charge_vars['amount']
  payto=charge_vars['charge_payee']
  payto_name=charge_vars['charge_payee_name']
  desc=charge_vars['description']
  admin_pin=get_admin_pin() 
  actives=read_active() # list of active ac_id's
  if(interactive): 
    print(str(len(actives))+' accounts to be charged')
    print('configuration to be used: '+str(charge_vars)) 
    print('In stepped mode you can view and approve each charge in turn.')
    print(' in non-stepped mode all charges will be added without') 
    print(' further prompting.')
    mode=input('Do you want to run in stepped mode [y/n] ?')
  else:
    mode='n'
  if mode == 'n' or mode == 'N':
    stepped=False
  else:
    stepped=True
  recno=0
  for acrec in actives:
    recno+=1
    proc='n'
    postdata=prepare_post(recno,admin_pin,domain,webapp_path,
            acrec['payer'],acrec['payer_name'],payto,payto_name,amount,desc)
    if stepped:
      print('record '+str(recno))
      print(str(acrec))
      print('\n'+str(postdata)+'\n')
      proc=input('proceed ? [y/n]')
    if not stepped or proc == 'y' or proc == 'Y': 
      do_post(postdata)

if __name__ == '__main__':
  main()

def test():
  ''' tests some of the utility functions in this module. To carry
  out a full integration test, this program has to be run from a 
  test system instance
  e.g. with 2 active accounts for Bob and Alice, see the testsuite. '''
  #   
  ap=get_admin_pin()
  print('admin pin: '+ap)
  # 
  actives=read_active()
  print('active accounts: ' + str(actives))
  # 
  notcharged=nochargeacs()
  print('uncharged accounts: ' + str(notcharged))
  #
  print(' site variables: ' + str(site_vars))
  print(' charge variables: ' + str(charge_vars))
