#!/usr/bin/env python

# change next line from 0 to 1 to debug script

debug=0

""" tcategories.cgi This file handles input from the 
    form used to create/delete/view/modify trading categories.
    Richard Kay partly based on input from Cathy Wattebot,
    Last changed  2023

"""


if debug:
  import sys
  sys.stderr=sys.stdout
  print("Content-type: text/plain\n")

import cgiutils
import webinput
woi=webinput.Webinput()
import tablcgi
from table_details import wotab,cattab,mtab
site_vars=cgiutils.get_vars('site.cfg')
dl=site_vars['data_location']
sec_vars=cgiutils.get_vars('secure.cfg')
 
def main():
  session=False
  cgiutils.html_header(title="LETS Trading Category Management",
                       bgcolor='"#FFFFFF"')
  valid,ac_id,pincookie,ua=cgiutils.get_hidden(woi)
  if not valid:
    return # error message already sent
  valid,option=cgiutils.get_option(woi,buttons=[
      "add","modify","delete","view"])
  if not valid:
    return # error message already sent
  # check pin_or_cookie and account
  session=cgiutils.login(ac_id,pincookie,ua)
  valid=session.login()
  cookie=session.cookie
  if not valid:
    cgiutils.html_header(title="LETS Accounting and Directory System")
    cgiutils.html_end(session,error="Invalid or stale or logged out session.")
    return
  # 
  # only do work for an admin account
  if ua != 'admin': 
    cgiutils.html_end(session,error="Administrator login required.")
    return
  
  # If we've got this far, there should be a valid option and PIN OK
  categs=tablcgi.table(cattab,dir=dl)
  if option == "add": # need details for new category 
    if not woi.has_required(["category","sort","desc"]):
      cgiutils.html_end(session,
        error="Missing field for add category. Complete all fields.")
      return
    # validate fields
    category=woi.firstval("category")
    if not cgiutils.is_valid(category,max_leng=15,
        allowed='^[A-Za-z]+$'):
      cgiutils.html_end(session,
         error="Invalid category. Use letters a-z only.")
      return
    sort=woi.firstval("sort")
    if not cgiutils.is_valid(sort,max_leng=4,
        allowed='^[0-9]+$'):
      cgiutils.html_end(session,
         error="Invalid sort code. Use numbers 0-9 only.")
      return
    desc=woi.firstval("desc")
    if not cgiutils.is_valid(desc,max_leng=60,
        allowed=r'''^[\w \'\-\.\,\!\"\'\$\%\^\&\*\(\)\+\=\;\:\?\/\\\@\|]+$'''):
      cgiutils.html_end(session,error="Invalid category description.")
      return
 
    # check category doesn't exist
    if category in categs.keys():
      cgiutils.html_end(session,error="Category exists. Can't add it.")
      return

    # if still here add row to categories table
    row={}
    row["category"]=category
    row["sort"]=int(sort)
    row["desc"]=desc
    if not cgiutils.auth(session,ac_id,"CATEGADD",cgiutils.mystr(row)):
      return
    categs.addrow(row)
    print("<p>new category added")
    cgiutils.html_end(session,received=1)
    return
  elif option == "modify": # modify details for category 
    category=woi.firstval("category")
    modifications=0 # counts number of fields changed
    # check account does exist
    if category not in categs.keys():
      error="Category doesn't exist. Can't modify."
      cgiutils.html_end(session,error)
      return
    else: # get category row index 
      category=woi.firstval("category")
      cindex=categs.find(category)

    # check sort code 
    if "sort" in woi.keys():
      # validate field
      sort=woi.firstval("sort")
      if not cgiutils.is_valid(sort,max_leng=4,allowed='^[0-9]+$'):
        cgiutils.html_end(session,
           error="Invalid sort code. Use numbers 0-9 only.")
        return
      categs.data[cindex]["sort"]=int(sort)
      modifications+=1
    # check description 
    if "desc" in woi.keys():
      # validate field
      desc=woi.firstval("desc")
      if not cgiutils.is_valid(desc,max_leng=60,
         allowed=r'''^[\w \'\-\.\,\!\"\'\$\%\^\&\*\(\)\+\=\;\:\?\/\\\@\|]+$'''):
        cgiutils.html_end(session,
           error="Invalid category description.")
        return
      categs.data[cindex]["desc"]=desc
      modifications+=1
    if modifications: # save changes to members table if any
      if not cgiutils.auth(session,ac_id,"CATEGMOD",str(categs.data[cindex])):
        return 
      categs.save()
      print("<p> %d field/s modified.</p>" % modifications)
      cgiutils.html_end(session,received=1)
      return
    else: # all changeable fields were blank 
      cgiutils.html_end(session,error="No modified field values submitted.")
      return
  elif option == "delete":
    category=woi.firstval("category")
    if category not in categs.keys():
      error="Category doesn't exist. Can't delete it."
      cgiutils.html_end(session,error)
      return
    cindex=categs.find(category)
    # best not to delete category if any wants and offers use it
    wantoffers=tablcgi.table(wotab,dir=dl)
    for worow in wantoffers.data:
      if worow["categ"] == category:
        error="Category currently used in want/offer table. Can't delete it."
        cgiutils.html_end(session,error)
        return
    if not cgiutils.auth(session,ac_id,"CATEGDEL",category):
      return
    categs.delrow(category)
    cgiutils.html_end(session,received=1)
    return
  elif option == "view":
    # view category table
    import table_details
    categs.sort(cmpfunc=table_details.cmp_categories)
    categs.tab2html()
    cgiutils.html_end(session)
    return

if __name__ == "__main__":
  try:
      main()
  except:
    import traceback
    print("error detected in tcategories.cgi main()")
    traceback.print_exc()

