#!/usr/bin/env python

''' Renumbers a set of Python tests sequentially starting at 1.
    A testsuite Python test source file e.g. testNN.py 
    comments tests as follows:

    # test 1:
    # test 2: 

    These numbers need maintenance whenever tests are inserted 
    or removed, and this program provides that maintenance. 

    This program renames the files in expected_results to correspond
    with new testnumbers,  without overwriting any existing
    expected result files. But doing this safely requires that
    new tests should be inserted with numbers not already
    present, e.g. if existing numbers end at 30 start numbering
    new tests at 40  
'''

import re
import os, os.path

expdir='expected_results'
if not os.path.isdir(expdir):
  raise ValueError("Current directory not testsuite or no expected_results")
# temp directory
tmpdir=os.path.join(expdir,'tmp')

def renumber(commentstring,newnum,testcase):
  ''' returns commentstring with newnum in place of 
  previous value. Also renumbers expected_result file 
  into tmp directory if one exists for old testnumber. '''
  cre=re.compile(r'(\# test )([0-9]+)(\:)')
  mo=cre.search(commentstring)
  if not mo:
    raise ValueError("incorrect comment")
  if not len(mo.groups()) == 3:
    raise ValueError("incorrect comment")
  startstr=commentstring[:mo.span()[0]]+mo.groups()[0]
  oldnum=int(mo.groups()[1])
  endstr=mo.groups()[2]+commentstring[mo.span()[1]:]
  middle=str(newnum)
  # compute expected_results filenames
  oldexpbn='test'+str(testcase)+'_'+str(oldnum)+'.tst' # basename 
  oldexpfn=os.path.join(expdir,oldexpbn) # fullname
  newexpbn='test'+str(testcase)+'_'+str(newnum)+'.tst' # basename
  newexpfn=os.path.join(tmpdir,newexpbn) # fullname
  if os.path.isfile(oldexpfn):
    print("renaming %s to %s" % (oldexpfn,newexpfn))   
    os.rename(oldexpfn,newexpfn)
  return startstr+middle+endstr

def replace_expected_results():
  ''' moves files renamed/renumbered from the expected_results 
  into tmp directory back into expected_results with new numbering '''
  count=0
  for bname in os.listdir(tmpdir): 
    fromname=os.path.join(tmpdir,bname) # fullname
    toname=os.path.join(expdir,bname) # fullname
    os.rename(fromname,toname)
    count+=1
  if count:
    print('%d expected_results files renumbered' % count)

def has_testnumber_comment(sourceline):
  ''' returns true if sourceline has test number comment '''
  cre=re.compile(r'\# test [0-9]+\:')
  mo=cre.search(sourceline)
  if mo:
    return True
  else:
    return False

def has_testcase(sourceline):
  ''' returns testcase assignment number or False '''
  cre=re.compile(r'testcase\=([0-9]+)')
  mo=cre.search(sourceline)
  if mo:
    tcnum=int(mo.groups()[0])
    if tcnum < 1:
      raise ValueError("testcase numbered less than 1") 
    return tcnum # testcase= assignment
  else:
    return False # doesn't have testcase= assignment

def main():
  if not os.path.isdir(tmpdir):
    os.makedirs(tmpdir,0o770)
  source=input("enter name of sourcefile: ") 
  fh=open(source,'r')
  lines=fh.readlines()
  fh.close()
  opsource=""
  newnum=1
  tstcase=0
  for line in lines:
    tc=has_testcase(line)
    if tc:
      testcase=tc # store testcase once found, other tc's False
      print('testcase: %d' % testcase)
    if not has_testnumber_comment(line):
      opsource+=line
    else:
      opsource+=renumber(line,newnum,testcase)
      newnum+=1
  ofh=open(source,'w')
  ofh.write(opsource)
  replace_expected_results() # put files from tmpdir back

if __name__ == '__main__':
  main()
