#!/usr/bin/env python
''' testlib.py Test module for Python CGI regression test scripts 

e.g. a user test module might do the following: 

import testlib
import re
v=testlib.get_vars()
# various statements concerned with creating the tests list
testlib.run_tests(tests)

Facilities for normalised comparisons of actual and expected file-based test
result data are provided. Normalisation of comparison data is 
required where this data stores IP addresses, dates and times which are 
inherently unpredictable. Fields in such cases are identified and standardised
using matching regular expressions, substituting a actual date/time fields or
IP address for strings such as: DATE/TIME or IPV4_ADDRESS. 
 
For comparisons of files to be a valid test, any surrogate keys must also
be predictable, requiring a clean system before obtaining correct
file or running test suite requiring this kind of test, i.e. 
using make_clean.py script. This test suite must
not be run on a live system. 

'''

#import urllib.request, urllib.error, urllib.parse
import requests 
import sys,os,os.path
sys.path.append('..') # parent directory contains cgiutils and tablcgi
import cgiutils, tablcgi, authent, table_details

def get_form(formname,vars=[],ua='user'):
  ''' uses the cgiutils.send_form function to obtain any form.
  The form is returned with or without admin includes depending
  upon the ua option. '''
  import cgiutils, tablcgi, authent, table_details, LockFile
  #olddir=os.getcwd()
  #os.chdir('..') # change to parent directory
  form=cgiutils.send_form(formname,vars,ua=ua,form_dir='../forms')  
  #os.chdir(olddir) # change back 
  return form

def ackfiles_select():
  ''' knowing the accounting form requires the select statement '''
  import cgiutils, tablcgi, authent, table_details, LockFile
  olddir=os.getcwd()
  os.chdir('..') # change to parent directory
  # sys.path.append('.') # probably need local directory files 
  # clone atab so we can change file and not poison it for other uses
  ackfiles=cgiutils.list_ackfiles()
  os.chdir(olddir) # change back 
  return cgiutils.make_file_select(ackfiles,
         instruction="choose current file or archive",
         cgiparameter="ackfile")

def get_vars(cfile='site.cfg',savedpin=None):
  ''' reads the test input variables, from a test config file '''
  config_location=cgiutils.cfg_directory
  fh=open(os.path.join(config_location,cfile)) # test configuration file
  vars={}
  lines=fh.readlines()
  fh.close()
  for line in lines:
    if line[0] == '#' or line[0] == ' ':
      continue # skip comments or empty lines
    name,value=line.split(':',1)
    if value[-1] == '\n':
      value=value[:-1] # strip newline from end
    vars[name]=value
  if savedpin:
    # read saved testing value for admin PIN 
    apinfh=open('admin_pin.tst')
    admin_pin=apinfh.read()
    apinfh.close()
    vars['pin']=admin_pin
  return vars

def qtodict(querys):
  ''' turns query part of url into dictionary '''
  d={}
  if querys == '':
      return d
  if querys[0] == '?':
      querys=querys[1:] # strip leading ?
  pairs=querys.split('&')
  for pair in pairs:
      parts=pair.split('=')
      if len(parts) < 2:
          next # ignore malformed parts
      name,value = (parts[0],parts[1])
      d[name] = value
  return d

def run_make_clean(v=None,tv=None):
  ''' runs remote make_clean.cgi to clean test system and
  obtain and update admin PIN'''
  import re
  if not v:
    v=get_vars()
  if not tv:
    tv=get_vars(cfile='test.cfg')
  parms={'make_clean':'make_clean','testing':'testing'}
  testurl='http://'+tv['domain']+tv['testdir']+'/test.cgi'
  req = requests.get(testurl,params=parms)
  result=req.text
  # extract admin pin from result
  import re
  cre=re.compile(r'([1-9]\d{7,7})') # 8 digit PIN between 1000 and 9999
  mo=cre.search(result)
  if not mo:
    raise ValueError("problem running make_clean on remote system result returned: "
                    +str(testurl)  
                     +str(parms)
                     +str(result) 
                     )
  else:
    admin_pin=mo.group(1)
    # save admin_pin to file for later test scripts in series which
    # follow on from previous test scripts which renew the pin
    apinfh=open('admin_pin.tst','w')
    apinfh.write(admin_pin)
    apinfh.close()
    return admin_pin 

def get_nextackid(tv=None):
  ''' runs remote getackid.cgi to obtain next ackid'''
  if not tv:
    tv=get_vars(cfile='test.cfg')
  testurl='http://'+tv['domain']+tv['testdir']+'test.cgi'
  querys='get_nextackid=get_nextackid'
  parmd=qtodict(querys)
  req = requests.get(testurl,params=parmd)
  result=req.text
  next_ackid=int(result)
  return next_ackid

def get_userpin(user,tv=None):
  ''' returns PIN for user ''' 
  if not tv:
    tv=get_vars(cfile='test.cfg')
  testurl='http://'+tv['domain']+tv['testdir']+'test.cgi'
  querys='get_userpin=get_userpin&ac_id=%s' % user
  parmd=qtodict(querys)
  req = requests.get(testurl,params=parmd)
  result=req.text
  # find table detail containing 8 digit PIN between 10000000 and 99999999
  import re
  cre=re.compile(r'([1-9][0-9]{7,7})')
  mo=cre.search(result)
  userpin=mo.group(1) # get 8 digit PIN as string 
  return userpin 

def get_confirmcode(email,tv=None):
  ''' returns PIN for user '''
  if not tv:
    tv=get_vars(cfile='test.cfg')
  testurl='http://'+tv['domain']+tv['testdir']+'test.cgi'
  querys='get_confirmcode=get_confirmcode&email=%s' % email 
  parmd=qtodict(querys)
  req = requests.get(testurl,params=parmd)
  result=req.text
  # find table detail containing 8 digit PIN between 1000000000 and 9999999999
  import re
  cre=re.compile(r'([1-9][0-9]{9,9})')
  mo=cre.search(result)
  confcode=mo.group(1) # get 10 digit PIN as string 
  return confcode 

def hack_ackdates(secs,ackfile=None):
  ''' Removes secs from dates of all acks. This hack is 
  required for testing year end/carry forward processing.
  Dont even think about doing this on a production system. 
  '''
  import sys
  sys.path.append('..') # parent directory contains cgiutils and tablcgi
  import cgiutils, tablcgi, LockFile
  site_vars=cgiutils.get_vars('site.cfg')
  dl=site_vars['data_location']
  import os
  import os.path
  olddir=os.getcwd()
  from table_details import atab 
  os.chdir('..') # change to parent directory
  sys.path.append('.') # probably need local directory files 
  # clone atab so we can change file and not poison it for other uses
  catab=tablcgi.table_def(atab)
  if ackfile:
    fullackfile=os.path.join(dl,ackfile)
    if not os.path.isfile(fullackfile):
      raise ValueError("cant hack dates on non-existent file")
    catab.file=ackfile
  else:
    catab.file=atab.file[:]
  acks=tablcgi.table(catab,dir=dl)
  for ackr in acks.data :
    ackr["time"]-=secs # hack date/time back by secs seconds
  acks.save() 
  os.chdir(olddir) # in case this function gets called from within module

def get_datestr(secs=0,format="%d %b %Y"):
  ''' returns date in string format. secs is added to 
  the current date/time. for format see the 
  documentation for time.strftime. 
  Default secs 0 is for now. default format outputs date as
  DD Mmm YYYYY e.g. 21 Apr 2009
  ''' 
  import time
  now=time.time()
  now_and_secs=now+secs 
  time_struct=time.localtime(now_and_secs) 
  strdate=time.strftime(format,time_struct) # string e.g. 21 Apr 2009
  return strdate

def logmtce(offset=5):
  ''' manipulates values in smaint table then 
  calls authent.maintenance() to rotate
  login logfiles and rebuild good/bad tables. Offset
  is the number of seconds that this function will
  make maintenance overdue by. '''
  import sys
  sys.path.append('..') # parent directory contains cgiutils and tablcgi
  import cgiutils, tablcgi
  site_vars=cgiutils.get_vars('site.cfg')
  dl=site_vars['data_location']

  import os
  olddir=os.getcwd()
  import time
  now=time.time()
  from table_details import smaint
  os.chdir('..') # change to parent directory
  sys.path.append('.') # authent needs a local directory file 
  import authent
  sm=tablcgi.table(smaint,dir=dl)
  newschedule={"lrotate":now-(3600*24*30),"nrotate":now-offset,
                "lgbr":now-(3600*24),"ngbr":now-offset}
  if not sm.data:
    sm.data.append(newschedule)
  else:
    sm.data[0]=newschedule
  sm.save()
  authent.maintenance()
  os.chdir(olddir)

def do_test(testdata,stepped=False):
  ''' perform a single test operation, by doing a HTTP post 
  operation on a CGI server, and reporting pass/fail .
  This is based upon whether the result string
  contains expected string (for a pass) or not (for a fail).

  This handles the simple case where direct comparison of an
  included expected string is present within the actual result.
  A set of normalisation functions are used to normalise
  expected and actual results prior to comparison, e.g. to
  convert uncertain dates and email addresses into string 
  constants describing the data type.
 
  For more complex test output checking the expected result must 
  be stored as a file in the expected_results directory and 
  the do_fb_test function is used for normalised comparison.
  '''
  testurl=testdata['url']
  querys=data=testdata['data']
  parmd=qtodict(querys)
  req = requests.get(testurl,params=parmd)
  result=req.text
  result=do_normalisations(result)
  expected=testdata['expected']
  expected=do_normalisations(expected)
  if result.find(expected) >= 0:
    # test passes if expected is substring of result
    if stepped:
      print('actual: ' + result)
      print('passed test: '+str(testdata['testno'])+' '+testdata['description'])
    return 'pass'
  else:
    print('failed test: '+str(testdata['testno'])+' '+testdata['description'])
    print('expected: ' + testdata['expected'])
    print('actual: ' + result)
    return 'fail'

def normalise_datetime(input):
  # following regex designed to match any date/time in particular format 
  # DOW MOY DD HH:MM:SS example:
  # Thu Aug 19 19:52:21 2004

  rex=r'((\w{3,3}\s{1,2}){2,2}\d{1,2}\s(\d\d\:){2,2}\d\d\s\d\d\d\d)'
  import re
  cre=re.compile(rex)
  subtext="DATETIME"
  mo=cre.search(input)
  while mo: # substitute all found date/time strings with subtext
    input=re.sub(cre,subtext,input)
    mo=cre.search(input)
  return input

def normalise_date(input):
  # following regex designed to match any date in particular format 
  # DOM MOY DD example:
  # 19 Aug 2004

  rex=r'[0123][0-9]\ (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\ 2[01]\d\d'
  import re
  cre=re.compile(rex)
  subtext="DATE"
  mo=cre.search(input)
  while mo: # substitute all found date strings with subtext
    input=re.sub(cre,subtext,input)
    mo=cre.search(input)
  return input

def normalise_ip4addr(input):
  # following regex designed to match any ip v4 address in dot quad format 
  # NN.NN.NN.NN example:
  # 234.0.12.154 

  rex=r'(\d{1,3}\.){3,3}\d{1,3}'
  import re
  cre=re.compile(rex)
  subtext="IPV4_OR_V6_ADDRESS"
  mo=cre.search(input)
  while mo: # substitute all found date strings with subtext
    input=re.sub(cre,subtext,input)
    mo=cre.search(input)
  return input

def normalise_ip6addr(input):
    """normalises IPv6 addresses.
       Regex source code for this function was obtained and copied from: 
       http://stackoverflow.com/questions/319279/how-to-validate-ip-address-in-python
       on Friday 16th Sept 2011 by R Kay. 
    """
    import re
    cre = re.compile(r"""
        #        \s*                         # Leading whitespace
        (?!.*::.*::)                # Only a single whildcard allowed
        (?:(?!:)|:(?=:))            # Colon iff it would be part of a wildcard
        (?:                         # Repeat 6 times:
            [0-9a-f]{0,4}           #   A group of at most four hexadecimal digits
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
        ){6}                        #
        (?:                         # Either
            [0-9a-f]{0,4}           #   Another group
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
            [0-9a-f]{0,4}           #   Last group
            (?: (?<=::)             #   Colon iff preceeded by exacly one colon
             |  (?<!:)              #
             |  (?<=:) (?<!::) :    #
             )                      # OR
         |                          #   A v4 address with NO leading zeros 
            (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            (?: \.
                (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            ){3}
        )
        # \s*                         # Trailing whitespace
    """, re.VERBOSE | re.IGNORECASE | re.DOTALL)
    subtext="IPV4_OR_V6_ADDRESS"
    mo=cre.search(input)
    while mo: # substitute all found date strings with subtext
      input=re.sub(cre,subtext,input)
      mo=cre.search(input)
    return input

def normalise_multiaddress(input):
  ''' Dirty hack which patches bug in normalise_ip6addr() resulting in a single IPV6 address being
      detected and substituted as an adjacent pair. Replaces paired IPV4_OR_V6_ADDRESSIPV4_OR_V6_ADDRESS 
      substitution with a single IPV4_OR_V6_ADDRESS string. Yes I know, much better to fix the buggy regex,
      but I don't know how as I didn't write it. '''
  rex=r'IPV4_OR_V6_ADDRESSIPV4_OR_V6_ADDRESS'
  import re
  cre=re.compile(rex)
  subtext="IPV4_OR_V6_ADDRESS"
  mo=cre.search(input)
  while mo: # substitute all found double address placeholder strings with subtext
    input=re.sub(cre,subtext,input)
    mo=cre.search(input)
  return input

def normalise_pintd(input):
  # following regex designed to match variable pin in <td> tag 
  # NNNNNNNN example:
  # 12345678 

  rex=r'\<td\>\s*\d{8,8}\s*\<\/td\>'
  import re
  cre=re.compile(rex,re.I)
  subtext="VARIABLE_PIN_TAG"
  mo=cre.search(input)
  while mo: # substitute all found pin tag strings with subtext
    input=re.sub(cre,subtext,input)
    mo=cre.search(input)
  return input

def normalise_cookie(input):
  # following regex designed to match variable cookie 
  # N+_N{12,12} example:
  # 123_123456789012 
  rex=r'\d+\_\d{12,12}'
  import re
  cre=re.compile(rex,re.I)
  subtext="VARIABLE_COOKIE"
  mo=cre.search(input)
  while mo: # substitute all found pin tag strings with subtext
    input=re.sub(cre,subtext,input)
    mo=cre.search(input)
  return input

def normalise_pinth(input):
  # following regex designed to match variable 12 hex pin hash digest in <td> tag 
  # NNNN example:
  # ab14f9e2f834 

  rex=r'\<td\>\s*[0-9a-fA-F]{12,12}\s*\<\/td\>'
  import re
  cre=re.compile(rex,re.I)
  subtext="VARIABLE_PIN_HDIG"
  mo=cre.search(input)
  while mo: # substitute all found pin tag strings with subtext
    input=re.sub(cre,subtext,input)
    mo=cre.search(input)
  return input

def normalise_sitevars(input):
  # string module used to find/replace site variables, e.g. email 
  # addresses in site.cfg to allow normalised comparison 
  # example:
  # admin@example.com 
  import sys
  sys.path.append('..') # testsuite parent directory contains cgiutils
  import cgiutils
  import os.path
  svars=cgiutils.get_vars('site.cfg')
  tvars=cgiutils.get_vars('test.cfg')
  allvars={} 
  # merge svars and tvars dictionaries
  for key in list(svars.keys()):
    allvars[key]=svars[key] 
  for key in list(tvars.keys()):
    allvars[key]=tvars[key] 
  # Only replace site variables in following list
  # add new ones as required. When adding always
  # put the more specific variables e.g. test_email1
  # before more general ones, e.g. email_domain.
  keylist=[ 'abuse', 'admin_e', 'homepage', 'test_email1', 
            'test_email2','email_domain' ]
  first=input
  second=first # needed in case no substitutions occur
  for key in keylist:
    if key not in allvars:
      raise ValueError("normalise_sitevars coding error")
    value=allvars[key] 
    second=first.replace(value,'SITE_CFG_VARIABLE_'+key.upper())
    first=second
  return second 

def do_normalisations(input):
  ''' performs series of normalisations on input string
  to enable fair comparison between an expected and actual
  test result. This function is called once for expected
  and once for actual for each test comparison.

  This job is done by registering and calling a sequence
  of functions which perform individual normalisations.
  '''
  nfuncs=[
    normalise_datetime,
    normalise_date,
    normalise_ip4addr,
    normalise_ip6addr,
    normalise_multiaddress,
    normalise_sitevars,
    normalise_pintd,
    normalise_cookie,
    normalise_pinth
  ]
  for func in nfuncs: # perform normalisations in above sequence
    input=func(input)
  return input

def do_fb_test(testdata,stepped=False,change_expected=False):
  ''' Do a File-Based Date-IPaddr-Email-PIN-Normalised test.
  This performs a single test operation, by doing a HTTP post
  operation on a CGI server, and reporting pass/fail .
  This is based on a comparison of result against a file resulting from
  correct operation from a previous test run, but with unknown fields
  including PINs, dates, date-times, and IP addresses 
  matching regexes having been normalised prior to comparison.

  If running in change_expected mode it provides an option to replace
  the expected result file with the actual result, but this
  needs to be used with caution to avoid incorrect results being 
  treated as OK and being subsequently expected. To make the job
  of comparison of small changes in large files easier a differences
  view is available. The actual result must be
  compared against realistic expectations about output from correct 
  program first before considering this correct from the POV of future 
  testing. 
  '''
  # do the test 
  testurl=testdata['url']
  querys=data=testdata['data']
  parmd=qtodict(querys)
  req = requests.get(testurl,params=parmd)
  result=req.text
  #req = urllib.request.Request(url=testdata['url'],data=testdata['data'])
  #fr = urllib.request.urlopen(req)
   
  # read actual and expected results 
  # result=fr.read()
  import os.path
  expected_directory='expected_results'
  expected_filename=os.path.join(expected_directory,testdata['resultfile'])
  efh=open(expected_filename)
  expected=efh.read()
  efh.close() 
  
  # normalise expected and actual results
  n_result=do_normalisations(result)
  n_expected=do_normalisations(expected)
  if n_result.find(n_expected) >= 0:
    # test passes if expected is substring of result
    if stepped:
      print('actual: ' + result)
    print('passed test: '+str(testdata['testno'])+' '+testdata['description'])
    return 'pass'
  else:
    print('failed test: '+str(testdata['testno'])+' '+testdata['description'])
    print('expected:\n\n ' + n_expected)
    print('\n\nactual: ' + n_result)
    actual_file=os.path.join(expected_directory,'actual.tst')
    afh=open(actual_file,'w')
    afh.write(n_result)
    afh.close()
    print('actual result saved to: '+actual_file)
    # The change_expected facility allows future expectations
    # to be updated in line with an actual result - use very carefully.
    if change_expected:
      import os, os.path
      expected=os.path.join('expected_results',testdata['resultfile'])
      actual=os.path.join('expected_results','actual.tst')
      option='r' # something not q
      while option in ["r","d","t"]: # menu loop
        print("you may now enter")
        print("r to replace expected result file with actual result")
        print("d to view differences")
        print("t to review testdata input for test")
        print("q to quit") 
        option=input("enter option: ").lower()[0]
        if option == 'r':
          if 'resultfile' not in testdata:
            print('cant replace if resultfile unspecified in test data')
            raise AttributeError("no resultfile specified in test data")
          else:
            os.rename(actual,expected)
            print('resultfile: '+testdata['resultfile']+' replaced')
            return 'pass'
        elif option == 'd':
          diffcmd="diff "+expected+" "+actual
          os.system(diffcmd)
        elif option == 't':
          print()
          print(str(testdata))
          print()
        elif option == 'q':
          return 'fail'
        else:
          print("invalid option, try again")
          option='r' # something which causes loop to continue 
    return 'fail'

def run_tests(tests,stepped=False,change_expected=False):
  ''' runs a series of tests, supplied as a list of 
  dictionaries, each dictionary containing the URL to be
  queried, the test number within the series, and information
  about the expected result of the test, supplied either as
  a string or filehandle. stepped parameter results in
  more generous output.
  '''
  import types
  for test in range(len(tests)):
    if 'expected' in list(tests[test].keys()):
      result=do_test(tests[test],stepped)
    elif 'resultfile' in list(tests[test].keys()):
      result=do_fb_test(tests[test],stepped,change_expected)
    else:
      raise Exception("missing data",
        "invalid missing test data expected or resultfile attribute")
    if result == 'pass':
      print("Test no: "+str(tests[test]["testno"])+" "+tests[test]["description"]+" passed")
    if result == 'fail':
      print("test data: "+str(tests[test]))
      print("\n\n retry url:"+tests[test]["url"]+"?"+tests[test]["data"]+"\n\n")
      raise Exception("fail","regression test failure - suite discontinued")
    if stepped:
      # provide breakpoint
      input("breakpoint: press enter to continue")

def modify_config_var(filename,varname,newval):
  ''' rewrites a variable value in a 
  testing .cfg file to enable different configuration options to be 
  tested. This is needed if the variable controls program logic, 
  to check if the user is authorised to do something.
 
  Parameters: filename - name of configuration file 
              varname - name of variable to be substituted
              newval - new value for variable
  '''
  # regex to substitute value for config_dir
  # portability note: might need to handle windows pathnames as well as Unix
  # example of a matching line: ip4addr:1.2.3.4
  regex=r'^'+varname+r'\:.+$' 
  cfh=open(filename,'r')
  cff=cfh.read() # read config file as string
  cfh.close()
  import re
  cre=re.compile(regex,re.MULTILINE)
  subtext=varname+':'+newval
  mo=cre.search(cff)
  if mo: # substitute found instances with subtext
    cff=re.sub(cre,subtext,cff)
  else:
    raise ValueError("Can't substitute invalid contents of "+filename)
  cfh=open(filename,'w')
  cfh.write(cff)
  cfh.close()
